# Copyright (c) 2022 ING Wholesale Banking Advanced Analytics
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import histogrammar as hg
import numpy as np

from ..analysis.hist_numpy import assert_similar_hists
from ..base import Module


class HistStitcher(Module):
    """Module stitches histograms by date"""

    _input_keys = ("read_key", "delta_key")
    _output_keys = ("store_key",)

    def __init__(
        self,
        mode="add",
        time_axis=None,
        time_bin_idx=None,
        read_key=None,
        delta_key=None,
        store_key=None,
    ):
        """Stitching histograms by first axis.

        :param str mode: options for histogram stitching: "add" or "replace". default is "add".
        :param str time_axis: name of the first axis, to stitch on.
        :param str time_bin_idx: value of delta dataset used for stitching,
            in case delta or first dataset is a batch without time_axis. Should be an ordered string or integer.
        :param str read_key: key of input histogram-dict to read from data store.
            (only required when calling transform(datastore) as module)
        :param str delta_key: key of delta histogram-dict to read from data store.
            (only required when calling transform(datastore) as module)
        :param str store_key: key of output data to store in data store
            (only required when calling transform(datastore) as module)
        """
        super().__init__()
        self.read_key = read_key
        self.delta_key = delta_key
        self.store_key = store_key
        self.mode = mode
        self.time_axis = time_axis
        self.time_bin_idx = time_bin_idx
        self.allowed_modes = ["add", "replace"]
        if self.mode not in self.allowed_modes:
            raise ValueError("mode should be either 'add' or 'replace'")

    def get_description(self):
        return f"{self.mode}"

    def transform(self, hists_basis: dict, hists_delta: dict) -> dict:
        self.logger.info(
            f'Stitching histograms "{self.read_key}" and "{self.delta_key}" as "{self.store_key}"'
        )
        stitched = self.stitch_histograms(self.mode, hists_basis, hists_delta)
        return stitched

    def stitch_histograms(
        self,
        mode=None,
        hists_basis=None,
        hists_delta=None,
        hists_list=None,
        time_axis="",
        time_bin_idx=None,
    ):
        """Stitching histograms by first axis.

        Histograms in hists_delta are added to those in hists_basis. Bins are summed or replaced, set this with 'mode'.
        'time_axis' specifies the name of the first axis. If the time_axis is not found, it is created, and
        histograms get inserted the time_bin_idx values.

        :param str mode: options for histogram stitching: "add" or "replace" bins. default is "add".
        :param dict hists_basis: input dict of basis histograms.
        :param dict hists_delta: delta dict of histograms to add to hists_basis.
        :param list hists_list: alternative for [hists_basis, hists_delta, etc]. can have multiple deltas.
            first item in list is taken as hists_basis (optional)
        :param str time_axis: time-axis used for stitching (optional).
        :param time_bin_idx: (list of) time-value(s) at which to insert hist-deltas into hist-basis.
        :return: dict with stitched histograms. If stitching is not possible, returns hists_basis.
        """
        # set stitching mode
        mode = (
            mode if isinstance(mode, str) and mode in self.allowed_modes else self.mode
        )
        time_axis = (
            time_axis
            if isinstance(time_axis, str) and len(time_axis) > 0
            else self.time_axis
        )
        time_bin_idx = (
            time_bin_idx
            if isinstance(time_bin_idx, (str, int, list, tuple))
            else self.time_bin_idx
        )
        hists_list = hists_list or []

        if time_bin_idx is not None:
            if isinstance(time_bin_idx, (str, int)):
                time_bin_idx = [time_bin_idx]
            if not isinstance(time_bin_idx, (list, tuple)):
                raise TypeError(
                    "time_bin_idx should be a (list of) ordered integer(s) or string(s)"
                )
            dts = [type(tv) for tv in time_bin_idx]
            if not dts.count(dts[0]) == len(dts):
                raise TypeError(f"time_bin_idxs have inconsistent datatypes: {dts}")

        # basic checks and conversions
        if isinstance(hists_basis, dict) and len(hists_basis) > 0:
            hists_list.insert(0, hists_basis)
        if isinstance(hists_delta, dict) and len(hists_delta) > 0:
            hists_list.append(hists_delta)
        elif isinstance(hists_delta, list):
            for hd in hists_delta:
                if isinstance(hd, dict) and len(hd) > 0:
                    hists_list.append(hd)
        if not isinstance(hists_list, list) or len(hists_list) == 0:
            raise TypeError(
                "hists_list should be a filled list of histogram dictionaries."
            )
        for hd in hists_list:
            if not isinstance(hd, dict) or len(hd) == 0:
                raise TypeError(
                    "hists_list should be a list of filled histogram dictionaries."
                )
        hists_basis = hists_list[0]
        hists_delta = hists_list[1:]

        # determine possible features, used for comparisons below
        if isinstance(time_axis, str) and len(time_axis) > 0:
            self.feature_begins_with = f"{time_axis}:"

        # Three possibilities
        # 1. if there are no basis hists starting with "time_axis:", assume that this the very first batch.
        # 2. if delta(s) do not start with "time_axis:", assume that this is a set of batches without time_axis
        # 3. deltas already have a time-axis. start stitching of histograms by using bins.update()

        features_basis = self.get_features(
            list(hists_basis.keys())
        )  # basis keys that start with time_axis
        if len(features_basis) > 0:
            # pruning of keys not starting with time_axis
            hists_list[0] = hists_basis = {
                k: h for k, h in hists_basis.items() if k in features_basis
            }

        # 1. if there are no basis hists starting with "time_axis:", assume that this the very first batch.
        if (
            len(features_basis) == 0
            and time_axis
            and len(hists_basis) > 0
            and time_axis
        ):
            if time_bin_idx is None:
                self.logger.info(
                    f'Inserting basis histograms in axis "{time_axis}" at bin index 0.'
                )
                time_bin_idx = [0]
            hists_basis_new = {}
            for k, hist in hists_basis.items():
                feature = f"{time_axis}:{k}"
                self.logger.debug(f'Now creating histogram "{feature}"')
                hists_basis_new[feature] = self._create_hist_with_time_axis(
                    hist, time_bin_idx[0]
                )
            # reset hists_basis
            features_basis = self.get_features(list(hists_basis_new.keys()))
            hists_list[0] = hists_basis = hists_basis_new
            time_bin_idx = time_bin_idx[1:]

        # -- is there a need to continue? There need to be overlapping hists.
        delta_keys = set()
        if len(hists_delta) > 0:
            delta_keys = set(hists_delta[0].keys())
            for hd in hists_delta[1:]:
                delta_keys &= set(hd.keys())
        if len(hists_delta) == 0 or len(delta_keys) == 0 or len(hists_basis) == 0:
            self.logger.debug(
                "No overlapping delta features. returning pruned hists_basis."
            )
            return hists_basis

        stitched = {}

        # 2. if delta(s) do not start with "time_axis:", assume that this is a set of batches without time_axis
        features_delta = self.get_features(
            list(delta_keys)
        )  # delta keys that start with time_axis
        if (
            len(features_basis) > 0
            and len(features_delta) == 0
            and len(delta_keys) > 0
            and time_axis
        ):
            if time_bin_idx is None or len(time_bin_idx) == 0:
                time_bin_idx = self._generate_time_bin_idx(
                    hists_basis, features_basis, time_axis, len(hists_delta)
                )
                if time_bin_idx is None:
                    raise ValueError(
                        "Request to insert delta hists but time_bin_idx not set or deductable. Please set manually."
                    )
                self.logger.info(
                    f'Inserting delta histograms in axis "{time_axis}" at bin indices {time_bin_idx}.'
                )
            if len(hists_delta) != len(time_bin_idx):
                raise ValueError(
                    "Not enough time_bin_idxs set to insert delta histograms."
                )
            for key in list(delta_keys):
                feature = f"{time_axis}:{key}"
                if feature not in features_basis:
                    continue
                self.logger.debug(f'Now inserting into histogram "{feature}"')
                hist_list = [hd[key] for hd in hists_delta]
                stitched[feature] = self._insert_hists(
                    hists_basis[feature], hist_list, time_bin_idx, mode
                )
            # add basis hists without any overlap
            for feature in features_basis:
                if feature not in stitched:
                    stitched[feature] = hists_basis[feature]
            return stitched

        # 3. deltas already have a time-axis. start stitching of histograms by using bins.update()
        overlapping_keys = set(hists_list[0].keys())
        for hd in hists_list[1:]:
            overlapping_keys &= set(hd.keys())
        features_overlap = self.get_features(
            list(overlapping_keys)
        )  # all overlapping keys that start with time_axis
        if len(features_overlap) == 0:
            # no overlap, then return basis histograms
            self.logger.warning(
                "No overlapping basis-delta features. returning pruned hists_basis."
            )
            return hists_basis
        for feature in features_overlap:
            self.logger.debug(f'Now stitching histograms "{feature}"')
            hist_list = [hd[feature] for hd in hists_list]
            stitched[feature] = self._stitch_by_update(mode, hist_list)
        # add basis hists without any overlap
        for feature in features_basis:
            if feature not in stitched:
                stitched[feature] = hists_basis[feature]
        return stitched

    def _find_max_time_bin_index(self, hists_basis, features_basis, time_axis):
        """Find the maximum time-bin index in dict of basis histograms

        :param hists_basis: dict of basis histograms
        :param features_basis: list of features to look at
        :param time_axis: name of time-axis
        :return: maximum time-bin index found or None
        """
        # basic checks
        assert isinstance(time_axis, str) and len(time_axis) > 0
        assert len(features_basis) > 0
        assert all(f.startswith(time_axis) for f in features_basis)

        hist_list = list(hists_basis.values())

        all_sparse = all(isinstance(h, hg.SparselyBin) for h in hist_list)
        all_cat = not all_sparse and all(
            isinstance(h, hg.Categorize) for h in hist_list
        )

        max_time_bin_idx = None
        if all_sparse or all_cat:
            max_time_bin_idx = max(
                max(h.bins.keys()) for h in hist_list if len(h.bins) > 0
            )
        return max_time_bin_idx

    def _generate_time_bin_idx(self, hists_basis, features_basis, time_axis, n):
        """Generate n consecutive time-bin indices beyond max existing time-bin index

        :param hists_basis: dict of basis histograms
        :param features_basis: list of features to look at
        :param time_axis: name of time-axis
        :param n: number of time-bin indices to generate
        :return: maximum time-bin index found or None
        """
        assert n > 0
        max_time_bin_idx = self._find_max_time_bin_index(
            hists_basis, features_basis, time_axis
        )
        if max_time_bin_idx is not None:
            self.logger.info(f"Maximum time bin index found: {max_time_bin_idx}")
        time_bin_idx = None
        if isinstance(max_time_bin_idx, (int, np.integer)):
            start = max_time_bin_idx + 1
            stop = start + n
            time_bin_idx = np.arange(start, stop)
        return time_bin_idx

    def _insert_hists(self, hbasis, hdelta_list, time_bin_idx, mode):
        """Add delta hist to hist_basis at time-value

        :param hbasis: basis histogram with time-axis
        :param hdelta_list: delta histogram without time-axis
        :param time_bin_idx: list of time-value at which to insert hdeltas into hbasis
        :param str mode: options for histogram stitching: "add" or "replace". default is "add".
        :return: stitched hbasis histogram
        """
        # basic checks on time-values
        if len(hdelta_list) != len(time_bin_idx) or len(hdelta_list) == 0:
            raise ValueError(
                "hdelta_list and time_bin_idx should be filled and same size."
            )
        dts = [type(tv) for tv in time_bin_idx]
        if not dts.count(dts[0]) == len(dts):
            raise TypeError(f"time_bin_idxs have inconsistent datatypes: {dts}")
        if not isinstance(time_bin_idx[0], (str, int, np.integer)):
            raise TypeError("time_bin_idxs should be an (ordered) string or integer.")

        # consistency checks on histogram definitions
        if not hasattr(hbasis, "bins"):
            raise ValueError(
                "basis histogram does not have bins attribute. cannot insert."
            )
        if len(hbasis.bins) > 0:
            hbk0 = list(hbasis.bins.values())[0]
            assert_similar_hists([hbk0] + hdelta_list)
        else:
            assert_similar_hists(hdelta_list)

        # check consistency of hbasis and time-value type
        if isinstance(time_bin_idx[0], str):
            if not isinstance(hbasis, hg.Categorize):
                raise TypeError("hbasis does not accept string time-values.")
        elif isinstance(time_bin_idx[0], (int, np.integer)):
            if not isinstance(hbasis, hg.SparselyBin):
                raise TypeError("hbasis does not accept integer time-values.")

        # stitch all the hdeltas into hbasis
        hsum = hbasis.copy()

        for tv, hdelta in zip(time_bin_idx, hdelta_list):
            if tv in hsum.bins:
                # replace or sum?
                hbt = hsum.bins[tv]
                hbt_entries = hbt.entries
                if mode == "replace":
                    hsum.bins[tv] = hdelta
                    hsum.entries += hdelta.entries - hbt_entries
                else:
                    hsum.bins[tv] += hdelta
                    hsum.entries += hdelta.entries
            else:
                # insert at time_bin_idx
                hsum.bins[tv] = hdelta
                hsum.entries += hdelta.entries

        return hsum

    def _create_hist_with_time_axis(self, hist, time_bin_idx):
        """Create histogram with time-axis and place hist into it at time-value

        :param hist: input histogram to insert into histogram with time-axis
        :param str time_bin_idx: time-value at which to insert histogram
        :return: histogram with time-axis
        """
        # basic checks
        if time_bin_idx is None or not isinstance(time_bin_idx, (str, int)):
            raise TypeError(
                "time_bin_idx not set. should be an (ordered) string or integer."
            )

        ht = (
            hg.SparselyBin(binWidth=1.0, origin=0.0, quantity=lambda x: x)
            if isinstance(time_bin_idx, int)
            else hg.Categorize(quantity=lambda x: x)
        )  # noqa
        ht.bins[time_bin_idx] = hist
        ht.entries = hist.entries
        return ht

    def _stitch_by_update(self, mode, hist_list):
        """Get sum of histograms using h1.bins.update(h2.bins), from first to last hist

        Get sum of list of histograms by using bins.update(), where the update is applied
        by iterating over the histogram list, from the first to the last histogram.
        Meaning the next bins can overwrite the previous bins.
        This is used for merging (bins of) histograms that can be topped up over time.

        :param str mode: options for histogram stitching: "add" or "replace". default is "add".
        :param list hist_list: list of input histogrammar histograms
        :return: list of consistent 1d numpy arrays with bin_entries for list of input histograms
        """
        # --- basic checks
        if len(hist_list) == 0:
            raise ValueError("Input histogram list has zero length.")
        assert_similar_hists(hist_list)

        # --- loop over all histograms and update zeroed-original consecutively
        hsum = hist_list[0].zero()

        if mode == "replace":
            # check consistency of histogram bins attributes
            is_bins = all(hasattr(hist, "bins") for hist in hist_list)
            if not is_bins:
                raise TypeError("Not all input histograms have bins attribute.")
            # update bins consecutively for each time-delta.
            for hist in hist_list:
                hsum.bins.update(hist.bins)
            hsum.entries = sum(b.entries for b in hsum.bins.values())
        else:
            for hist in hist_list:
                hsum += hist
        return hsum


def stitch_histograms(
    mode=None,
    hists_basis=None,
    hists_delta=None,
    hists_list=None,
    time_axis=None,
    time_bin_idx=None,
):
    """Stitching histograms by first axis.

    Histograms in hists_delta are added to those in hists_basis. Bins are summed or replaced, set this with 'mode'.
    'time_axis' specifies the name of the first axis. If the time_axis is not found, it is created, and
    histograms get inserted the time_bin_idx values.

    :param str mode: options for histogram stitching: "add" or "replace" bins. default is "add".
    :param dict hists_basis: input dict of basis histograms.
    :param dict hists_delta: delta dict of histograms to add to hists_basis.
    :param list hists_list: alternative for [hists_basis, hists_delta, etc]. can have multiple deltas.
        first item in list is taken as hists_basis (optional)
    :param str time_axis: time-axis used for stitching (optional).
    :param time_bin_idx: (list of) time-value(s) at which to insert hist-deltas into hist-basis.
    :return: dict with stitched histograms. If stitching is not possible, returns hists_basis.
    """
    stitcher = HistStitcher()
    return stitcher.stitch_histograms(
        mode=mode,
        hists_basis=hists_basis,
        hists_delta=hists_delta,
        hists_list=hists_list,
        time_axis=time_axis,
        time_bin_idx=time_bin_idx,
    )



# Copyright (c) 2022 ING Wholesale Banking Advanced Analytics
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


from pathlib import Path
from typing import Callable, Optional, Union

from ..base import Module


class FileReader(Module):
    """Module to read contents from a file, transform the contents with a function and write them to the datastore."""

    _input_keys = ()
    _output_keys = ("store_key",)

    def __init__(
        self,
        store_key: str,
        file_path: Union[str, Path],
        apply_func: Optional[Callable] = None,
        **kwargs,
    ):
        """Initialize an instance.

        :param str store_key: key of input data to be stored in the datastore
        :param str file_path: the file path to read the data from
        :param callable apply_func: function to be used for the transformation of data (optional)
        :param dict kwargs: additional keyword arguments which would be passed to `apply_func`
        """
        super().__init__()
        if not isinstance(file_path, (str, Path)):
            raise TypeError("file_path should be of type `str` or `pathlib.Path`")
        if apply_func is not None and not callable(apply_func):
            raise TypeError("transformation function must be a callable object")

        self.store_key = store_key
        self.file_path = file_path
        self.apply_func = apply_func
        self.kwargs = kwargs

    def get_description(self):
        return self.file_path

    def transform(self):
        with open(self.file_path) as file:
            data = file.read()

        # if a transformation function is provided, transform the data
        if self.apply_func is not None:
            data = self.apply_func(data, **self.kwargs)

        self.logger.info(
            f'Object "{self.store_key}" read from file "{self.file_path}".'
        )

        # store the transformed/original contents
        return data



# Copyright (c) 2022 ING Wholesale Banking Advanced Analytics
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import numpy as np
import pandas as pd
from scipy.stats import norm, pearsonr

from ...analysis.apply_func import ApplyFunc
from ...analysis.functions import (
    expand_norm_hist_mean_cov,
    expanding_hist,
    hist_sum,
    normalized_hist_mean_cov,
    relative_chi_squared,
    roll_norm_hist_mean_cov,
    rolling_hist,
)
from ...analysis.hist_numpy import (
    check_similar_hists,
    get_consistent_numpy_1dhists,
    get_consistent_numpy_entries,
    get_consistent_numpy_ndgrids,
)
from ...base import Pipeline
from ...hist.hist_utils import COMMON_HIST_TYPES, is_numeric
from ...stats.numpy import ks_prob, ks_test, uu_chi2


def hist_compare(row, hist_name1="", hist_name2="", max_res_bound=7.0):
    """Function to compare two histograms

    Apply statistical tests to compare two input histograms, such as:
    Chi2, KS, Pearson, max probability difference.
    For categorical histograms, also check for unknown labels.

    :param pd.Series row: row to apply compare function to
    :param str hist_name1: name of histogram one to compare
    :param str hist_name2: name of histogram two to compare
    :param float max_res_bound: count number of normalized residuals with (absolute) value greater than X.
        Default is 7.0.
    :return: pandas Series with popular comparison metrics.
    """
    from .comparisons import Comparisons

    x = {
        "ks": np.nan,
        "ks_zscore": np.nan,
        "ks_pvalue": np.nan,
        "pearson": np.nan,
        "chi2": np.nan,
        "chi2_norm": np.nan,
        "chi2_zscore": np.nan,
        "chi2_pvalue": np.nan,
        "chi2_max_residual": np.nan,
        "chi2_spike_count": np.nan,
        "unknown_labels": np.nan,
    }

    for key in Comparisons.get_comparisons().keys():
        x[key] = np.nan

    # basic name checks
    cols = row.index.to_list()
    if len(hist_name1) == 0 or len(hist_name2) == 0 and len(cols) == 2:
        hist_name1 = cols[0]
        hist_name2 = cols[1]
    if not all(name in cols for name in [hist_name1, hist_name2]):
        raise ValueError("Need to provide two histogram column names.")

    # basic histogram checks
    hist1 = row[hist_name1]
    hist2 = row[hist_name2]
    if not all(
        isinstance(hist, COMMON_HIST_TYPES) for hist in [hist1, hist2]
    ) or not check_similar_hists([hist1, hist2]):
        return pd.Series(x)

    # compare
    if hist1.n_dim == 1:
        if is_numeric(hist1):
            numpy_1dhists = get_consistent_numpy_1dhists([hist1, hist2])
            entries_list = [nphist[0] for nphist in numpy_1dhists]
            # KS-test only properly defined for (ordered) 1D interval variables
            ks_testscore = ks_test(*entries_list)
            x["ks"] = ks_testscore
            ks_pvalue = ks_prob(ks_testscore)
            x["ks_pvalue"] = ks_pvalue
            x["ks_zscore"] = -norm.ppf(ks_pvalue)
        else:  # categorical
            entries_list = get_consistent_numpy_entries([hist1, hist2])
            # check consistency of bin_labels
            labels1 = hist1.keySet
            labels2 = hist2.keySet
            subset = labels1 <= labels2
            x["unknown_labels"] = int(not subset)
    else:
        numpy_ndgrids = get_consistent_numpy_ndgrids([hist1, hist2], dim=hist1.n_dim)
        entries_list = [entry.flatten() for entry in numpy_ndgrids]

    # calculate pearson coefficient
    pearson, pvalue = (np.nan, np.nan)
    if len(entries_list[0]) >= 2:
        same0 = all(entries_list[0] == entries_list[0][0])
        same1 = all(entries_list[1] == entries_list[1][0])
        if not same0 and not same1:
            # this avoids std==0, and thereby avoid runtime warnings
            pearson, pvalue = pearsonr(*entries_list)

    chi2, chi2_norm, zscore, pvalue, res = uu_chi2(*entries_list)
    abs_residual = np.abs(res)
    chi2_max_residual = np.max(abs_residual)
    chi2_spike_count = np.sum(abs_residual[abs_residual > max_res_bound])

    x["pearson"] = pearson
    x["chi2"] = chi2
    x["chi2_norm"] = chi2_norm
    x["chi2_zscore"] = zscore
    x["chi2_pvalue"] = pvalue
    x["chi2_max_residual"] = chi2_max_residual
    x["chi2_spike_count"] = chi2_spike_count
    for key, func in Comparisons.get_comparisons().items():
        x[key] = func(*entries_list)
    return pd.Series(x)


class HistComparer(Pipeline):
    """Base pipeline to compare histogram to previous rolling histograms"""

    def __init__(
        self,
        func_hist_collector,
        read_key,
        store_key,
        assign_to_key=None,
        hist_col="histogram",
        suffix="comp",
        max_res_bound=7.0,
        *args,
        **kwargs,
    ):
        """Initialize an instance of RollingHistComparer.

        :param func_hist_collector: histogram collection function
        :param str read_key: key of input data to read from data store
        :param str store_key: key of output data to store in data store
        :param str assign_to_key: key of the input data to assign function applied-output to. (optional)
        :param str hist_col: column/key in input df/dict that contains the histogram. default is 'histogram'
        :param str suffix: column/key of rolling histogram. default is 'roll' -> column = 'histogram_roll'
        :param float max_res_bound: count number of normalized residuals with (absolute) value greater than X.
                                    Default is 7.0.
        :param args: (tuple, optional): residual args passed on to func_mean and func_std
        :param kwargs: (dict, optional): residual kwargs passed on to func_mean and func_std
        """
        if assign_to_key is None:
            assign_to_key = read_key

        # make reference histogram(s)
        hist_collector = ApplyFunc(
            apply_to_key=read_key,
            assign_to_key=assign_to_key,
        )
        hist_collector.add_apply_func(
            func=func_hist_collector, entire=True, suffix=suffix, *args, **kwargs
        )
        # do histogram comparison
        hist_comparer = ApplyFunc(
            apply_to_key=assign_to_key,
            assign_to_key=store_key,
            apply_funcs=[
                {
                    "func": hist_compare,
                    "hist_name1": hist_col,
                    "hist_name2": hist_col + "_" + suffix,
                    "prefix": suffix,
                    "axis": 1,
                    "max_res_bound": max_res_bound,
                }
            ],
        )

        super().__init__(modules=[hist_collector, hist_comparer])


class RollingHistComparer(HistComparer):
    """Compare histogram to previous rolling histograms"""

    def __init__(
        self,
        read_key,
        store_key,
        window,
        shift=1,
        hist_col="histogram",
        suffix="roll",
        max_res_bound=7.0,
    ):
        """Initialize an instance of RollingHistComparer.

        :param str read_key: key of input data to read from data store
        :param str store_key: key of output data to store in data store
        :param int window: size of rolling window
        :param int shift: shift of rolling window. default is 1.
        :param str hist_col: column/key in input df/dict that contains the histogram. default is 'histogram'
        :param str suffix: column/key of rolling histogram. default is 'roll' -> column = 'histogram_roll'
        :param float max_res_bound: count number of normalized residuals with (absolute) value greater than X.
                                    Default is 7.0.
        """
        super().__init__(
            rolling_hist,
            read_key,
            store_key,
            read_key,
            hist_col,
            suffix,
            max_res_bound,
            window=window,
            shift=shift,
            hist_name=hist_col,
        )
        self.read_key = read_key
        self.window = window

    def transform(self, datastore):
        self.logger.info(
            f'Comparing "{self.read_key}" with rolling sum of {self.window} previous histogram(s).'
        )
        return super().transform(datastore)


class PreviousHistComparer(RollingHistComparer):
    """Compare histogram to previous histograms"""

    def __init__(
        self,
        read_key,
        store_key,
        hist_col="histogram",
        suffix="prev1",
        max_res_bound=7.0,
    ):
        """Initialize an instance of PreviousHistComparer.

        :param str read_key: key of input data to read from data store
        :param str store_key: key of output data to store in data store
        :param str hist_col: column/key in input df/dict that contains the histogram. default is 'histogram'
        :param str suffix: column/key of rolling histogram. default is 'prev' -> column = 'histogram_prev'
        :param float max_res_bound: count number of normalized residuals with (absolute) value greater than X.
                                    Default is 7.0.
        """
        super().__init__(
            read_key,
            store_key,
            window=1,
            shift=1,
            hist_col=hist_col,
            suffix=suffix,
            max_res_bound=max_res_bound,
        )


class ExpandingHistComparer(HistComparer):
    """Compare histogram to previous expanding histograms"""

    def __init__(
        self,
        read_key,
        store_key,
        shift=1,
        hist_col="histogram",
        suffix="expanding",
        max_res_bound=7.0,
    ):
        """Initialize an instance of ExpandingHistComparer.

        :param str read_key: key of input data to read from data store
        :param str store_key: key of output data to store in data store
        :param int shift: shift of rolling window. default is 1.
        :param str hist_col: column/key in input df/dict that contains the histogram. default is 'histogram'
        :param str suffix: column/key of rolling histogram. default is 'expanding' -> column = 'histogram_expanding'
        :param float max_res_bound: count number of normalized residuals with (absolute) value greater than X.
                                    Default is 7.0.
        """
        super().__init__(
            expanding_hist,
            read_key,
            store_key,
            read_key,
            hist_col,
            suffix,
            max_res_bound,
            shift=shift,
            hist_name=hist_col,
        )
        self.read_key = read_key

    def transform(self, datastore):
        self.logger.info(
            f'Comparing "{self.read_key}" with expanding sum of past histograms.'
        )
        return super().transform(datastore)


class ReferenceHistComparer(HistComparer):
    """Compare histogram to reference histograms"""

    def __init__(
        self,
        reference_key,
        assign_to_key,
        store_key,
        hist_col="histogram",
        suffix="ref",
        max_res_bound=7.0,
    ):
        """Initialize an instance of ReferenceHistComparer.

        :param str reference_key: key of input data to read from data store
        :param str assign_to_key: key of input data to read from data store
        :param str store_key: key of output data to store in data store
        :param str hist_col: column/key in input df/dict that contains the histogram. default is 'histogram'
        :param str suffix: column/key of rolling histogram. default is 'ref' -> column = 'histogram_ref'
        :param float max_res_bound: count number of normalized residuals with (absolute) value greater than X.
                                    Default is 7.0.
        """
        super().__init__(
            hist_sum,
            reference_key,
            store_key,
            assign_to_key,
            hist_col,
            suffix,
            max_res_bound,
            metrics=[hist_col],
        )
        self.reference_key = reference_key
        self.assign_to_key = assign_to_key

    def transform(self, datastore):
        self.logger.info(
            f'Comparing "{self.assign_to_key}" with reference "{self.reference_key}"'
        )
        return super().transform(datastore)


class NormHistComparer(Pipeline):
    """Base pipeline to compare histogram to normalized histograms"""

    def __init__(
        self,
        func_hist_collector,
        read_key,
        store_key,
        assign_to_key=None,
        hist_col="histogram",
        *args,
        **kwargs,
    ):
        """Initialize an instance of NormHistComparer.

        :param func_hist_collector: histogram collection function
        :param str read_key: key of input data to read from data store
        :param str store_key: key of output data to store in data store
        :param str assign_to_key: key of the input data to assign function applied-output to. (optional)
        :param str hist_col: column/key in input df/dict that contains the histogram. default is 'histogram'
        :param args: (tuple, optional): residual args passed on to func_hist_collector
        :param kwargs: (dict, optional): residual kwargs passed on to func_hist_collector
        """
        if assign_to_key is None:
            assign_to_key = read_key

        # make reference histogram(s)
        hist_collector = ApplyFunc(apply_to_key=read_key, assign_to_key=assign_to_key)
        hist_collector.add_apply_func(
            func=func_hist_collector, hist_name=hist_col, suffix="", *args, **kwargs
        )

        # do histogram comparison
        hist_comparer = ApplyFunc(
            apply_to_key=assign_to_key,
            assign_to_key=store_key,
            apply_funcs=[
                {
                    "func": relative_chi_squared,
                    "hist_name": hist_col,
                    "suffix": "",
                    "axis": 1,
                }
            ],
        )

        super().__init__(modules=[hist_collector, hist_comparer])


class RollingNormHistComparer(NormHistComparer):
    """Compare histogram to previous rolling normalized histograms"""

    def __init__(self, read_key, store_key, window, shift=1, hist_col="histogram"):
        """Initialize an instance of RollingNormHistComparer.

        :param str read_key: key of input data to read from data store
        :param str store_key: key of output data to store in data store
        :param int window: size of rolling window
        :param int shift: shift of rolling window. default is 1.
        :param str hist_col: column/key in input df/dict that contains the histogram. default is 'histogram'
        """
        if window < 2:
            raise ValueError("Need window size of 2 or greater.")
        super().__init__(
            roll_norm_hist_mean_cov,
            read_key,
            store_key,
            read_key,
            hist_col,
            window=window,
            shift=shift,
            entire=True,
        )
        self.read_key = read_key
        self.window = window

    def transform(self, datastore):
        self.logger.info(
            f'Comparing "{self.read_key}" with relative mean of {self.window} previous histogram(s).'
        )
        return super().transform(datastore)


class ExpandingNormHistComparer(NormHistComparer):
    """Compare histogram to previous expanding normalized histograms"""

    def __init__(self, read_key, store_key, shift=1, hist_col="histogram"):
        """Initialize an instance of ExpandingNormHistComparer.

        :param str read_key: key of input data to read from data store
        :param str store_key: key of output data to store in data store
        :param int shift: shift of rolling window. default is 1.
        :param str hist_col: column/key in input df/dict that contains the histogram. default is 'histogram'
        """
        super().__init__(
            expand_norm_hist_mean_cov,
            read_key,
            store_key,
            read_key,
            hist_col,
            shift=shift,
            entire=True,
        )
        self.read_key = read_key

    def transform(self, datastore):
        self.logger.info(
            f'Comparing "{self.read_key}" with normalized mean of expanding past histograms.'
        )
        return super().transform(datastore)


class ReferenceNormHistComparer(NormHistComparer):
    """Compare histogram to reference normalized histograms"""

    def __init__(self, reference_key, assign_to_key, store_key, hist_col="histogram"):
        """Initialize an instance of ReferenceNormHistComparer.

        :param str reference_key: key of input data to read from data store
        :param str assign_to_key: key of input data to read from data store
        :param str store_key: key of output data to store in data store
        :param str hist_col: column/key in input df/dict that contains the histogram. default is 'histogram'
        """
        super().__init__(
            normalized_hist_mean_cov, reference_key, store_key, assign_to_key, hist_col
        )
        self.reference_key = reference_key
        self.assign_to_key = assign_to_key

    def transform(self, datastore):
        self.logger.info(
            f'Comparing "{self.assign_to_key}" with normalized reference "{self.reference_key}"'
        )
        return super().transform(datastore)



# Copyright (c) 2022 ING Wholesale Banking Advanced Analytics
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import numpy as np
import pandas as pd
from numpy.lib.stride_tricks import as_strided
from scipy import linalg, stats
from scipy.stats import linregress, norm

from ..analysis.hist_numpy import (
    check_similar_hists,
    get_consistent_numpy_2dgrids,
    get_consistent_numpy_entries,
    set_2dgrid,
)
from ..hist.hist_utils import COMMON_HIST_TYPES, is_numeric
from ..stats.numpy import probability_distribution_mean_covariance


def pull(row, suffix_mean="_mean", suffix_std="_std", cols=None):
    """Calculate normalized residual (pull) for list of cols

    Function can be used by ApplyFunc module.

    :param pd.Series row: row to apply pull function to
    :param list cols: list of cols to calculate pull of
    :param str suffix_mean: suffix of mean. mean column = metric + suffix_mean
    :param str suffix_std: suffix of std. std column = metric + suffix_std
    """

    rdict = row.to_dict()
    if cols is None or len(cols) == 0:
        # if no columns are given, find columns for which pulls can be calculated.
        # e.g. to calculate x_pull, need to have [x, x_mean, x_std] present. If so, put x in cols.
        indices = list(rdict.keys())
        cols = [
            m
            for m in set(indices)
            if all(r in indices for r in [m + suffix_mean, m + suffix_std])
        ]

    x = {
        m: np.nan
        if (
            any(
                r not in rdict or pd.isnull(rdict[r])
                for r in [m, m + suffix_mean, m + suffix_std]
            )
            or rdict[m + suffix_std] == 0.0
        )
        else (rdict[m] - rdict[m + suffix_mean]) / rdict[m + suffix_std]
        for m in cols
    }

    return pd.Series(x)


def expanding_mean(df, shift=1):
    """Calculate expanding mean of all numeric columns of a pandas dataframe

    Function can be used by ApplyFunc module.

    :param pd.DataFrame df: input pandas dataframe
    :param int shift: size of shift. default is 1.
    :return: df with expanding means of columns
    """
    return df.shift(shift).expanding().mean()


def expanding_std(df, shift=1):
    """Calculate expanding std of all numeric columns of a pandas dataframe

    Function can be used by ApplyFunc module.

    :param pd.DataFrame df: input pandas dataframe
    :param int shift: size of shift. default is 1.
    :return: df with expanding std of columns
    """
    return df.shift(shift).expanding().std()


def expanding_apply(df, func, shift=1, *args, **kwargs):
    """Calculate expanding apply() to all columns of a pandas dataframe

    Function can be used by ApplyFunc module.

    :param pd.DataFrame df: input pandas dataframe
    :param func: function to be applied
    :param int shift: size of shift. default is 1.
    :param args: args passed on to function
    :param kwargs: kwargs passed on to function
    :return: df with expanding results of function applied to all columns
    """
    return df.shift(shift).expanding().apply(func, args=args, **kwargs)


def rolling_std(df, window, shift=1):
    """Calculate rolling std of all numeric columns of a pandas dataframe

    Function can be used by ApplyFunc module.

    :param pd.DataFrame df: input pandas dataframe
    :param int shift: size of shift. default is 1.
    :param int window: size of rolling window.
    :return: df with rolling std of columns
    """
    return df.shift(shift).rolling(window).std()


def rolling_mean(df, window, shift=1):
    """Calculate rolling mean of all numeric columns of a pandas dataframe

    Function can be used by ApplyFunc module.

    :param pd.DataFrame df: input pandas dataframe
    :param int shift: size of shift. default is 1.
    :param int window: size of rolling window.
    :return: df with rolling mean of columns
    """
    return df.shift(shift).rolling(window).mean()


def rolling_apply(df, window, func, shift=1, *args, **kwargs):
    """Calculate rolling apply() to all columns of a pandas dataframe

    Function can be used by ApplyFunc module.

    :param pd.DataFrame df: input pandas dataframe
    :param int window: size of rolling window.
    :param func: function to be applied
    :param int shift: size of shift. default is 1.
    :param args: args passed on to function
    :param kwargs: kwargs passed on to function
    :return: df with rolling results of function applied to all columns
    """
    # raw=False already use Future setting
    return df.shift(shift).rolling(window).apply(func, raw=False, args=args, **kwargs)


def rolling_lr(df, window, index=0, shift=0):
    """Calculate rolling scipy lin_regress() to all columns of a pandas dataframe

    Function can be used by ApplyFunc module.

    :param pd.DataFrame df: input pandas dataframe
    :param int window: size of rolling window.
    :param int index: index of lin_regress results to return. default is 0.
    :param int shift: size of shift. default is 0.
    :return: df with rolling results of lin_regress() function applied to all columns
    """
    # raw=True suppresses Future warning
    return (
        df.shift(shift)
        .rolling(window)
        .apply(lambda x: linregress(np.arange(len(x)), x)[index], raw=True)
    )


def rolling_lr_zscore(df, window, shift=0):
    """Calculate rolling z-score of scipy lin_regress() to all columns of a pandas dataframe

    Function can be used by ApplyFunc module.

    :param pd.DataFrame df: input pandas dataframe
    :param int window: size of rolling window.
    :param int shift: size of shift. default is 0.
    :return: df with rolling z-score results of lin_regress() function applied to all columns
    """
    # MB 20200420: turn original df.rolling off, it doesn't accept timestamps.
    # raw=True suppresses Future warning
    # return df.shift(shift).rolling(window).apply(func, raw=True)
    def func(x):
        y = pd.Series(index=x.index, dtype=float)
        for name in x.index:
            try:
                xv = x[name].astype(float)
                y[name] = norm.ppf(linregress(np.arange(len(xv)), xv)[3])
            except Exception:
                y[name] = np.nan
        return y

    return roll(df, window=window, shift=shift).apply(func, axis=1)


def roll(df, window, shift=1):
    """Implementation of rolling window that can handle non-numerical columns such as histograms

    :param pd.DataFrame df: input dataframe to apply rolling function to.
    :param int window: size of rolling window
    :param int shift: shift of dataframe, default is 1 (optional)
    """
    assert shift >= 0
    assert isinstance(
        df, (pd.DataFrame, pd.Series)
    ), "input should be a dataframe or series"

    cols = df.columns if isinstance(df, pd.DataFrame) else [df.name]
    x = df.values
    # apply shift
    x = x[:-shift] if shift > 0 else x

    # apply windowing, use numpy's as_strided function to step through x and create sub-arrays
    if isinstance(df, pd.DataFrame):
        d0, d1 = x.shape
        s0, s1 = x.strides
        arr = as_strided(x, (d0 - (window - 1), window, d1), (s0, s0, s1))
    elif isinstance(df, pd.Series):
        hopsize = 1
        nrows = ((x.size - window) // hopsize) + 1
        if nrows < 0:
            nrows = 0
        n = x.strides[0]
        arr = as_strided(x, shape=(nrows, window), strides=(hopsize * n, n))

    # fill up missing values b/c off window & shift with Nones
    arr_shape = list(arr.shape)
    arr_shape[0] = len(df.index) - len(arr)
    arr_shape = tuple(arr_shape)
    n_fill = len(cols) * window * (len(df.index) - len(arr))
    fill_value = np.array([[None] * n_fill]).reshape(arr_shape)
    arr = np.append(fill_value, arr, axis=0)

    # reshape to new data frame
    def reshape(vs, i):
        return vs if len(vs.shape) == 1 else vs[:, i]

    d = [{c: reshape(vals, i) for i, c in enumerate(cols)} for vals in arr]
    rolled_df = pd.DataFrame(data=d, index=df.index)

    return rolled_df


def expand(df, shift=1):
    """Implementation of expanding window that can handle non-numerical values such as histograms

    Split up input array into expanding sub-arrays

    :param pd.DataFrame df: input dataframe to apply rolling function to.
    :param int shift: shift of dataframe, default is 1 (optional)
    :param fillvalue: default value to fill dataframe in case shift > 0 (optional)
    """
    assert shift >= 0
    assert isinstance(
        df, (pd.DataFrame, pd.Series)
    ), "input should be a dataframe or series"

    cols = df.columns if isinstance(df, pd.DataFrame) else [df.name]
    x = df.values
    arr = [x[: max(i + 1 - shift, 0)] for i in range(x.shape[0])]

    # fill up missing values b/c off shift with Nones
    fill_value = np.array([[None] * len(cols)])
    for i in range(shift):
        arr[i] = fill_value

    # reshape to new data frame
    def reshape(vs, i):
        return vs if len(vs.shape) == 1 else vs[:, i]

    d = [{c: reshape(vals, i) for i, c in enumerate(cols)} for vals in arr]
    expanded_df = pd.DataFrame(data=d, index=df.index)

    return expanded_df


def expanding_hist(df, shift=1, *args, **kwargs):
    """Apply expanding histogram sum

    Function can be used by ApplyFunc module.

    :param pd.DataFrame df: input pandas dataframe with column of histograms
    :param int shift: shift of dataframe, default is 1 (optional)
    :param args: args passed on to hist_sum function
    :param kwargs: kwargs passed on to hist_sum function
    :return: dataframe with expanding hist_sum results
    """
    return expand(df, shift=shift).apply(hist_sum, axis=1, args=args, **kwargs)


def rolling_hist(df, window, shift=1, *args, **kwargs):
    """Apply rolling histogram sum

    Function can be used by ApplyFunc module.

    :param pd.DataFrame df: input pandas dataframe with column of histograms
    :param int window: size of rolling window
    :param int shift: shift of dataframe, default is 1 (optional)
    :param args: args passed on to hist_sum function
    :param kwargs: kwargs passed on to hist_sum function
    :return: dataframe with rolling hist_sum results
    """
    return roll(df, window=window, shift=shift).apply(
        hist_sum, axis=1, args=args, **kwargs
    )


def hist_sum(x, hist_name=""):
    """Return sum of histograms

    Usage: df['hists'].apply(hist_sum) ; series.apply(hist_sum)

    :param pd.Series x: pandas series to extract histogram list from.
    :param str hist_name: name of column to extract histograms from. needs to be set with axis=1 (optional)
    :return: sum histogram
    """
    assert isinstance(x, pd.Series)
    if len(hist_name) > 0 and hist_name in x:
        hist_list = x[hist_name]
    else:
        hist_list = x.values
        if len(hist_name) == 0:
            hist_name = "histogram"

    if len(hist_list) == 0:
        raise ValueError("List of input histograms is empty.")

    # initialize
    o = {hist_name: None}

    # basic checks
    all_hist = all(isinstance(hist, COMMON_HIST_TYPES) for hist in hist_list)
    if not all_hist:
        return pd.Series(o)

    similar = check_similar_hists(hist_list)
    if not similar:
        return pd.Series(o)

    # MB FIX: h_sum not initialized correctly in a sum by histogrammar for sparselybin (origin); below it is.
    # h_sum = np.sum([hist for hist in hist_list])

    h_sum = hist_list[0].zero()
    for hist in hist_list:
        h_sum += hist
    o[hist_name] = h_sum
    return pd.Series(o)


def roll_norm_hist_mean_cov(df, window, shift=1, *args, **kwargs):
    """Apply rolling normalized_hist_mean_cov function

    Function can be used by ApplyFunc module.

    :param pd.DataFrame df: input pandas dataframe with column of histograms
    :param int window: size of rolling window
    :param int shift: shift of dataframe, default is 1 (optional)
    :param args: args passed on to hist_sum function
    :param kwargs: kwargs passed on to hist_sum function
    :return: dataframe with rolling normalized_hist_mean_cov results
    """
    return roll(df, window=window, shift=shift).apply(
        normalized_hist_mean_cov, axis=1, args=args, **kwargs
    )


def expand_norm_hist_mean_cov(df, shift=1, *args, **kwargs):
    """Apply expanding normalized_hist_mean_cov function

    Function can be used by ApplyFunc module.

    :param pd.DataFrame df: input pandas dataframe with column of histograms
    :param int shift: shift of dataframe, default is 1 (optional)
    :param args: args passed on to hist_sum function
    :param kwargs: kwargs passed on to hist_sum function
    :return: dataframe with expanding normalized_hist_mean_cov results
    """
    return expand(df, shift=shift).apply(
        normalized_hist_mean_cov, axis=1, args=args, **kwargs
    )


def normalized_hist_mean_cov(x, hist_name=""):
    """Mean normalized histogram and its covariance of list of input histograms

    Usage: df['hists'].apply(normalized_hist_mean_cov) ; series.apply(normalized_hist_mean_cov)

    :param pd.Series x: pandas series to extract histogram list from.
    :param str hist_name: name of column to extract histograms from. needs to be set with axis=1 (optional)
    :return: mean normalized histogram, covariance probability matrix
    """
    assert isinstance(x, pd.Series)
    if len(hist_name) > 0 and hist_name in x:
        hist_list = x[hist_name]
    else:
        hist_list = x.values
        if len(hist_name) == 0:
            hist_name = "histogram"

    if len(hist_list) == 0:
        raise ValueError("List of input histograms is empty.")

    # initialize
    o = {
        hist_name + "_mean": None,
        hist_name + "_cov": None,
        hist_name + "_binning": None,
    }

    # basic checks
    all_hist = all(isinstance(hist, COMMON_HIST_TYPES) for hist in hist_list)
    if not all_hist:
        return pd.Series(o)
    similar = check_similar_hists(hist_list)
    if not similar:
        return pd.Series(o)

    # get entries as numpy arrays
    if hist_list[0].n_dim == 1:
        entries_list, binning = get_consistent_numpy_entries(
            hist_list, get_bin_labels=True
        )
        entries_list = np.array(entries_list, dtype=np.float)
    else:
        entries_list, keys = get_consistent_numpy_2dgrids(
            hist_list, get_bin_labels=True
        )
        entries_list = np.array([h.flatten() for h in entries_list], dtype=np.float)
        binning = tuple(keys)

    # calculation of mean normalized histogram and its covariance matrix
    (
        normalized_hist_mean,
        normalized_hist_covariance,
    ) = probability_distribution_mean_covariance(entries_list)

    o[hist_name + "_mean"] = normalized_hist_mean
    o[hist_name + "_cov"] = normalized_hist_covariance
    o[hist_name + "_binning"] = binning
    return pd.Series(o)


def relative_chi_squared(
    row,
    hist_name="histogram",
    suffix_mean="_mean",
    suffix_cov="_cov",
    suffix_binning="_binning",
):
    """Calculate chi squared of normalized histogram with pre-calculated mean normalized histogram

    :param pd.Series row: row to apply chi_squared function to.
    :param str hist_name: name of column to extract histograms from. default is 'histogram' (optional)
    :param str suffix_mean: suffix of mean. mean column = hist_name + suffix_mean (optional)
    :param str suffix_std: suffix of std. std column = hist_name + suffix_std (optional)
    :param str suffix_binning: suffix of binning. binning column = hist_name + suffix_binning (optional)
    """
    x = {
        "chi2": np.nan,
        "naive_pvalue": np.nan,
        "naive_zscore": np.nan,
        "max_res": np.nan,
    }

    required = [
        hist_name,
        hist_name + suffix_mean,
        hist_name + suffix_cov,
        hist_name + suffix_binning,
    ]
    if not all(r in row for r in required):
        return pd.Series(x)

    hist = row[hist_name]
    norm_mean = row[hist_name + suffix_mean]
    cov = row[hist_name + suffix_cov]
    binning = row[hist_name + suffix_binning]

    # basic checks
    if not isinstance(hist, COMMON_HIST_TYPES):
        return pd.Series(x)
    if any(ho is None for ho in [norm_mean, cov, binning]):
        return pd.Series(x)
    if len(cov.shape) != 2 or len(norm_mean.shape) != 1:
        return pd.Series(x)

    variance = np.diagonal(cov)

    # get entries as numpy arrays
    if hist.n_dim == 1:
        entries = (
            hist.bin_entries(xvalues=binning)
            if is_numeric(hist)
            else hist.bin_entries(labels=binning)
        )
    else:
        assert len(binning) == 2
        entries = set_2dgrid(hist, [binning[0], binning[1]])
        entries = entries.flatten()

    # calculation of mean normalized histogram and its covariance matrix of input histogram
    single_norm, _ = probability_distribution_mean_covariance([entries])

    if (
        np.linalg.cond(cov) < 0.1 / np.finfo(cov.dtype).eps
        and np.abs(np.linalg.det(cov)) > np.finfo(cov.dtype).eps
    ):
        # check if covariance matrix is invertible
        # see: https://stackoverflow.com/questions/13249108/efficient-pythonic-check-for-singular-matrix
        # We try to use the precision matrix (inverse covariance matrix) for the chi-squared calculation
        pm = linalg.inv(cov)
        chi_squared = np.dot(
            (norm_mean - single_norm), np.dot(pm, (norm_mean - single_norm))
        )
        if chi_squared <= 0:
            chi_squared = np.finfo(np.float).eps
    else:
        # If a covariance matrix is singular we fall back on using variances
        chi_squared = np.sum(
            (norm_mean - single_norm) ** 2 / (variance + np.finfo(np.float).eps)
        )

    # pvalue and zvalue based on naive number of degrees of freedom
    ndof = len(entries) - 1
    p_value = stats.chi2.sf(chi_squared, ndof)
    z_score = -stats.norm.ppf(p_value)
    # scenario where pvalue is too small to evaluate Z
    # use Chernoff approximation for p-value upper bound
    # see: https://en.wikipedia.org/wiki/Chi-squared_distribution
    if p_value == 0:
        z = chi_squared / ndof
        u = -np.log(2 * np.pi) - ndof * np.log(z) + ndof * (z - 1)
        z_score = np.sqrt(u - np.log(u))

    max_resid = np.max(
        np.abs((norm_mean - single_norm) / np.sqrt(variance + np.finfo(np.float).eps))
    )

    x["chi2"] = chi_squared
    x["naive_pvalue"] = p_value
    x["naive_zscore"] = z_score
    x["max_res"] = max_resid
    return pd.Series(x)



# Copyright (c) 2022 ING Wholesale Banking Advanced Analytics
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


import logging

import pandas as pd
from histogrammar.dfinterface.make_histograms import (
    get_bin_specs,
    get_time_axes,
    make_histograms,
)

from ..pipeline.metrics_pipelines import create_metrics_pipeline

logging.basicConfig(
    level=logging.INFO, format="%(asctime)s %(levelname)s [%(module)s]: %(message)s"
)
logger = logging.getLogger()


def stability_metrics(
    hists,
    reference_type="self",
    reference=None,
    time_axis="",
    window=10,
    shift=1,
    monitoring_rules=None,
    pull_rules=None,
    features=None,
    **kwargs,
):
    """Create a data stability monitoring datastore for given dict of input histograms.

    :param dict hists: input histograms to be profiled and monitored over time.
    :param reference_type: type or reference used for comparisons. Options [self, external, rolling, expanding].
        default is 'self'.
    :param reference: histograms used as reference. default is None
    :param str time_axis: name of datetime feature, used as time axis, eg 'date'. auto-guessed when not provided.
    :param int window: size of rolling window and/or trend detection. default is 10.
    :param int shift: shift of time-bins in rolling/expanding window. default is 1.
    :param dict monitoring_rules: monitoring rules to generate traffic light alerts.
        The default setting is:

        .. code-block:: python

            monitoring_rules = {
                "*_pull": [7, 4, -4, -7],
                "*_zscore": [7, 4, -4, -7],
                "[!p]*_unknown_labels": [0.5, 0.5, 0, 0],
            }

        Note that the (filename based) wildcards such as * apply to all statistic names matching that pattern.
        For example, ``"*_pull"`` applies for all features to all statistics ending on "_pull".
        You can also specify rules for specific features and/or statistics by leaving out wildcard and putting the
        feature name in front. E.g.

        .. code-block:: python

            monitoring_rules = {
                "featureA:*_pull": [5, 3, -3, -5],
                "featureA:nan": [4, 1, 0, 0],
                "*_pull": [7, 4, -4, -7],
                "nan": [8, 1, 0, 0],
            }

        In case of multiple rules could apply for a feature's statistic, the most specific one applies.
        So in case of the statistic "nan": "featureA:nan" is used for "featureA", and the other "nan" rule
        for all other features.
    :param dict pull_rules: red and yellow (possibly dynamic) boundaries shown in plots in the report.
        Default is:

        .. code-block:: python

            pull_rules = {"*_pull": [7, 4, -4, -7]}

        This means that the shown yellow boundaries are at -4, +4 standard deviations around the (reference) mean,
        and the shown red boundaries are at -7, +7 standard deviations around the (reference) mean.
        Note that the (filename based) wildcards such as * apply to all statistic names matching that pattern.
        (The same string logic applies as for monitoring_rules.)
    :param list features: histograms to pick up from the 'hists' dictionary (default is all keys)
    :param kwargs: residual keyword arguments passed on to report pipeline.
    :return: dict with results of metrics pipeline
    """

    if not isinstance(hists, dict):
        raise TypeError("hists should be a dict of histogrammar histograms.")

    if not isinstance(monitoring_rules, dict):
        monitoring_rules = {
            "*_pull": [7, 4, -4, -7],
            "*_zscore": [7, 4, -4, -7],
            "[!p]*_unknown_labels": [0.5, 0.5, 0, 0],
        }
    if not isinstance(pull_rules, dict):
        pull_rules = {"*_pull": [7, 4, -4, -7]}

    if (isinstance(time_axis, str) and len(time_axis) == 0) or (
        isinstance(time_axis, bool) and time_axis
    ):
        # auto guess the time_axis: find the most frequent first column name in the histograms list
        first_cols = [k.split(":")[0] for k in list(hists.keys())]
        time_axis = max(set(first_cols), key=first_cols.count)

    pipeline = create_metrics_pipeline(
        reference_type=reference_type,
        reference=reference,
        hists_key="hists",
        ref_hists_key="ref_hists",
        time_axis=time_axis,
        window=window,
        shift=shift,
        monitoring_rules=monitoring_rules,
        pull_rules=pull_rules,
        features=features,
        **kwargs,
    )

    datastore = {"hists": hists}
    if reference_type == "external":
        datastore["ref_hists"] = reference

    return pipeline.transform(datastore)


def df_stability_metrics(
    df,
    time_axis,
    features=None,
    binning="auto",
    bin_specs=None,
    time_width=None,
    time_offset=0,
    var_dtype=None,
    reference_type="self",
    reference=None,
    window=10,
    shift=1,
    monitoring_rules=None,
    pull_rules=None,
    **kwargs,
):
    """Create a data stability monitoring html datastore for given pandas or spark dataframe.

    :param df: input pandas/spark dataframe to be profiled and monitored over time.
    :param str time_axis: name of datetime feature, used as time axis, eg 'date'. if True, will be auto-guessed.
        If time_axis is set or found, and if no features provided, features becomes: ['date:x', 'date:y', 'date:z'] etc.
    :param list features: columns to pick up from input data. (default is all features).
        For multi-dimensional histograms, separate the column names with a ':'. Example features list is:

        .. code-block:: python

            features = ["x", "date", "date:x", "date:y", "date:x:y"]

    :param str binning: default binning to revert to in case bin_specs not supplied. options are:
        "unit" or "auto", default is "auto". When using "auto", semi-clever binning is automatically done.
    :param dict bin_specs: dictionaries used for rebinning numeric or timestamp features.
        An example bin_specs dictionary is:

        .. code-block:: python

            bin_specs = {
                "x": {"bin_width": 1, "bin_offset": 0},
                "y": {"num": 10, "low": 0.0, "high": 2.0},
                "x:y": [{}, {"num": 5, "low": 0.0, "high": 1.0}],
            }

        In the bin specs for x:y, x is not provided (here) and reverts to the 1-dim setting.
        The 'bin_width', 'bin_offset' notation makes an open-ended histogram (for that feature) with given bin width
        and offset. The notation 'num', 'low', 'high' gives a fixed range histogram from 'low' to 'high' with 'num'
        number of bins.
    :param time_width: bin width of time axis. str or number (ns). note: bin_specs takes precedence. (optional)

        .. code-block:: text

            Examples: '1w', 3600e9 (number of ns),
                      anything understood by pd.Timedelta(time_width).value

    :param time_offset: bin offset of time axis. str or number (ns). note: bin_specs takes precedence. (optional)

        .. code-block:: text

            Examples: '1-1-2020', 0 (number of ns since 1-1-1970),
                      anything parsed by pd.Timestamp(time_offset).value

    :param dict var_dtype: dictionary with specified datatype per feature. auto-guessed when not provided.
    :param reference_type: type or reference used for comparisons. Options [self, external, rolling, expanding].
        default is 'self'.
    :param reference: reference dataframe or histograms. default is None
    :param int window: size of rolling window and/or trend detection. default is 10.
    :param int shift: shift of time-bins in rolling/expanding window. default is 1.
    :param dict monitoring_rules: monitoring rules to generate traffic light alerts.
        The default setting is:

        .. code-block:: python

            monitoring_rules = {
                "*_pull": [7, 4, -4, -7],
                "*_zscore": [7, 4, -4, -7],
                "[!p]*_unknown_labels": [0.5, 0.5, 0, 0],
            }

        Note that the (filename based) wildcards such as * apply to all statistic names matching that pattern.
        For example, ``"*_pull"`` applies for all features to all statistics ending on "_pull".
        You can also specify rules for specific features and/or statistics by leaving out wildcard and putting the
        feature name in front. E.g.

        .. code-block:: python

            monitoring_rules = {
                "featureA:*_pull": [5, 3, -3, -5],
                "featureA:nan": [4, 1, 0, 0],
                "*_pull": [7, 4, -4, -7],
                "nan": [8, 1, 0, 0],
            }

        In case of multiple rules could apply for a feature's statistic, the most specific one applies.
        So in case of the statistic "nan": "featureA:nan" is used for "featureA", and the other "nan" rule
        for all other features.
    :param dict pull_rules: red and yellow (possibly dynamic) boundaries shown in plots in the report.
        Default is:

        .. code-block:: python

            pull_rules = {"*_pull": [7, 4, -4, -7]}

        This means that the shown yellow boundaries are at -4, +4 standard deviations around the (reference) mean,
        and the shown red boundaries are at -7, +7 standard deviations around the (reference) mean.
        Note that the (filename based) wildcards such as * apply to all statistic names matching that pattern.
        (The same string logic applies as for monitoring_rules.)
    :param kwargs: residual keyword arguments, passed on to stability_report()
    :return: dict with results of metrics pipeline
    """
    # basic checks on presence of time_axis
    if not (isinstance(time_axis, str) and len(time_axis) > 0) and not (
        isinstance(time_axis, bool) and time_axis
    ):
        raise TypeError("time_axis needs to be a filled string or set to True")
    if isinstance(time_axis, str) and time_axis not in df.columns:
        raise ValueError(f'time_axis  "{time_axis}" not found in columns of dataframe.')
    if reference is not None and not isinstance(reference, dict):
        if isinstance(time_axis, str) and time_axis not in reference.columns:
            raise ValueError(
                f'time_axis  "{time_axis}" not found in columns of reference dataframe.'
            )
    if isinstance(time_axis, bool):
        time_axes = get_time_axes(df)
        num = len(time_axes)
        if num == 1:
            time_axis = time_axes[0]
            logger.info(f'Time-axis automatically set to "{time_axis}"')
        elif num == 0:
            raise ValueError(
                "No obvious time-axes found. Cannot generate stability report."
            )
        else:
            raise ValueError(
                f"Found {num} time-axes: {time_axes}. Set *one* time_axis manually!"
            )
    if features is not None:
        # by now time_axis is defined. ensure that all histograms start with it.
        if not isinstance(features, list):
            raise TypeError(
                "features should be list of columns (or combos) to pick up from input data."
            )
        features = [
            c if c.startswith(time_axis) else f"{time_axis}:{c}" for c in features
        ]

    # interpret time_width and time_offset
    if isinstance(time_width, (str, int, float)) and isinstance(
        time_offset, (str, int, float)
    ):
        if bin_specs is None:
            bin_specs = {}
        elif not isinstance(bin_specs, dict):
            raise TypeError("bin_specs object is not a dictionary")

        if time_axis in bin_specs:
            raise ValueError(
                f'time-axis "{time_axis}" already found in binning specifications.'
            )
        # convert time width and offset to nanoseconds
        time_specs = {
            "bin_width": float(pd.Timedelta(time_width).value),
            "bin_offset": float(pd.Timestamp(time_offset).value),
        }
        bin_specs[time_axis] = time_specs

    reference_hists = None
    if reference is not None:
        reference_type = "external"
        if isinstance(reference, dict):
            # 1. reference is dict of histograms
            # extract features and bin_specs from reference histograms
            reference_hists = reference
            features = list(reference_hists.keys())
            bin_specs = get_bin_specs(reference_hists)
        else:
            # 2. reference is pandas or spark dataframe
            # generate histograms and return updated features, bin_specs, time_axis, etc.
            (
                reference_hists,
                features,
                bin_specs,
                time_axis,
                var_dtype,
            ) = make_histograms(
                reference,
                features,
                binning,
                bin_specs,
                time_axis,
                var_dtype,
                ret_specs=True,
            )

    # use the same features, bin_specs, time_axis, etc as for reference hists
    hists = make_histograms(
        df,
        features=features,
        binning=binning,
        bin_specs=bin_specs,
        time_axis=time_axis,
        var_dtype=var_dtype,
    )

    # generate data stability report
    return stability_metrics(
        hists,
        reference_type,
        reference_hists,
        time_axis,
        window,
        shift,
        monitoring_rules,
        pull_rules,
        features,
        **kwargs,
    )



# Copyright (c) 2022 ING Wholesale Banking Advanced Analytics
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


from pathlib import Path

from ..base import Pipeline
from ..config import config
from ..io import FileWriter
from ..pipeline.metrics_pipelines import (
    ExpandingReferenceMetricsPipeline,
    ExternalReferenceMetricsPipeline,
    RollingReferenceMetricsPipeline,
    SelfReferenceMetricsPipeline,
)
from ..visualization import (
    AlertSectionGenerator,
    HistogramSection,
    ReportGenerator,
    SectionGenerator,
    TrafficLightSectionGenerator,
)
from ..visualization.overview_section import OverviewSectionGenerator


def get_report_pipeline_class(reference_type, reference):
    _report_pipeline = {
        "self": SelfReference,
        "external": ExternalReference,
        "rolling": RollingReference,
        "expanding": ExpandingReference,
    }
    reference_types = list(_report_pipeline.keys())
    if reference_type not in reference_types:
        raise ValueError(f"reference_type should be one of {str(reference_types)}.")
    if reference_type == "external" and not isinstance(reference, dict):
        raise TypeError("reference should be a dict of histogrammar histograms.")

    return _report_pipeline[reference_type]


class SelfReference(Pipeline):
    def __init__(
        self,
        hists_key="test_hists",
        time_axis="date",
        window=10,
        monitoring_rules={},
        pull_rules={},
        features=None,
        skip_empty_plots=True,
        last_n=0,
        top_n=20,
        plot_hist_n=6,
        report_filepath=None,
        show_stats=None,
        disable_heatmap=None,
        cmap=None,
        **kwargs,
    ):
        """Example pipeline for comparing test data with itself (full test set)

        :param str hists_key: key to test histograms in datastore. default is 'test_hists'
        :param str time_axis: name of datetime feature. default is 'date' (column should be timestamp, date(time) or numeric batch id)
        :param int window: window size for trend detection. default is 10
        :param dict monitoring_rules: traffic light rules
        :param dict pull_rules: pull rules to determine dynamic boundaries
        :param list features: features of histograms to pick up from input data (optional)
        :param bool skip_empty_plots: if false, also show empty plots in report with only nans or zeroes (optional)
        :param int last_n: plot statistic data for last 'n' periods (optional)
        :param int top_n: plot heatmap for top 'n' categories. default is 20 (optional)
        :param int plot_hist_n: plot histograms for last 'n' periods. default is 1 (optional)
        :param str report_filepath: the file path where to output the report (optional)
        :param list show_stats: list of statistic name patterns to show in the report. If None, show all (optional)
        :param list disable_heatmap: list of heatmap types to disable in the report. 'normal' to disable normal heatmap, 'row' to row normalized, 'column' to disable column normalized. If None, show all (optional)
        :param kwargs: residual keyword arguments
        :return: assembled self reference pipeline
        """
        modules = [
            SelfReferenceMetricsPipeline(
                hists_key,
                time_axis,
                window,
                monitoring_rules,
                pull_rules,
                features,
                **kwargs,
            ),
            ReportPipe(
                sections_key="report_sections",
                store_key="html_report",
                skip_empty_plots=skip_empty_plots,
                last_n=last_n,
                top_n=top_n,
                plot_hist_n=plot_hist_n,
                report_filepath=report_filepath,
                show_stats=show_stats,
                disable_heatmap=disable_heatmap,
                cmap=cmap,
            ),
        ]

        super().__init__(modules)


class ExternalReference(Pipeline):
    def __init__(
        self,
        hists_key="test_hists",
        ref_hists_key="ref_hists",
        time_axis="date",
        window=10,
        monitoring_rules={},
        pull_rules={},
        features=None,
        skip_empty_plots=True,
        last_n=0,
        top_n=20,
        plot_hist_n=2,
        report_filepath=None,
        show_stats=None,
        disable_heatmap=None,
        cmap=None,
        **kwargs,
    ):
        """Example pipeline for comparing test data with other (full) external reference set

        :param str hists_key: key to test histograms in datastore. default is 'test_hists'
        :param str ref_hists_key: key to reference histograms in datastore. default is 'ref_hists'
        :param str time_axis: name of datetime feature. default is 'date' (column should be timestamp, date(time) or numeric batch id)
        :param int window: window size for trend detection. default is 10
        :param dict monitoring_rules: traffic light rules
        :param dict pull_rules: pull rules to determine dynamic boundaries
        :param list features: features of histograms to pick up from input data (optional)
        :param bool skip_empty_plots: if false, show empty plots in report with only nans or zeroes (optional)
        :param int last_n: plot statistic data for last 'n' periods (optional)
        :param int top_n: plot heatmap for top 'n' categories. default is 20 (optional)
        :param int plot_hist_n: plot histograms for last 'n' periods. default is 1 (optional)
        :param str report_filepath: the file path where to output the report (optional)
        :param list show_stats: list of statistic name patterns to show in the report. If None, show all (optional)
        :param list disable_heatmap: list of heatmap types to disable in the report. 'normal' to disable normal heatmap, 'row' to row normalized, 'column' to disable column normalized. If None, show all (optional)
        :param kwargs: residual keyword arguments
        :return: assembled external reference pipeline
        """
        modules = [
            ExternalReferenceMetricsPipeline(
                hists_key,
                ref_hists_key,
                time_axis,
                window,
                monitoring_rules,
                pull_rules,
                features,
                **kwargs,
            ),
            ReportPipe(
                sections_key="report_sections",
                store_key="html_report",
                skip_empty_plots=skip_empty_plots,
                last_n=last_n,
                top_n=top_n,
                plot_hist_n=plot_hist_n,
                report_filepath=report_filepath,
                show_stats=show_stats,
                disable_heatmap=disable_heatmap,
                cmap=cmap,
            ),
        ]

        super().__init__(modules)


class RollingReference(Pipeline):
    def __init__(
        self,
        hists_key="test_hists",
        time_axis="date",
        window=10,
        shift=1,
        monitoring_rules={},
        pull_rules={},
        features=None,
        skip_empty_plots=True,
        last_n=0,
        top_n=20,
        plot_hist_n=6,
        report_filepath=None,
        show_stats=None,
        disable_heatmap=None,
        cmap=None,
        **kwargs,
    ):
        """Example pipeline for comparing test data with itself (rolling test set)

        :param str hists_key: key to test histograms in datastore. default is 'test_hists'
        :param str time_axis: name of datetime feature. default is 'date' (column should be timestamp, date(time) or numeric batch id)
        :param int window: size of rolling window and for trend detection. default is 10
        :param int shift: shift in rolling window. default is 1
        :param dict monitoring_rules: traffic light rules
        :param dict pull_rules: pull rules to determine dynamic boundaries
        :param list features: features of histograms to pick up from input data (optional)
        :param bool skip_empty_plots: if false, show empty plots in report with only nans or zeroes (optional)
        :param int last_n: plot statistic data for last 'n' periods (optional)
        :param int top_n: plot heatmap for top 'n' categories. default is 20 (optional)
        :param int plot_hist_n: plot histograms for last 'n' periods. default is 1 (optional)
        :param str report_filepath: the file path where to output the report (optional)
        :param list show_stats: list of statistic name patterns to show in the report. If None, show all (optional)
        :param list disable_heatmap: list of heatmap types to disable in the report. 'normal' to disable normal heatmap, 'row' to row normalized, 'column' to disable column normalized. If None, show all (optional)
        :param kwargs: residual keyword arguments
        :return: assembled rolling reference pipeline
        """
        modules = [
            RollingReferenceMetricsPipeline(
                hists_key,
                time_axis,
                window,
                shift,
                monitoring_rules,
                pull_rules,
                features,
                **kwargs,
            ),
            ReportPipe(
                sections_key="report_sections",
                store_key="html_report",
                skip_empty_plots=skip_empty_plots,
                last_n=last_n,
                top_n=top_n,
                plot_hist_n=plot_hist_n,
                report_filepath=report_filepath,
                show_stats=show_stats,
                disable_heatmap=disable_heatmap,
                cmap=cmap,
            ),
        ]

        super().__init__(modules)


class ExpandingReference(Pipeline):
    def __init__(
        self,
        hists_key="test_hists",
        time_axis="date",
        window=10,
        shift=1,
        monitoring_rules={},
        pull_rules={},
        features=None,
        skip_empty_plots=True,
        last_n=0,
        top_n=20,
        plot_hist_n=6,
        report_filepath=None,
        show_stats=None,
        disable_heatmap=None,
        cmap="autumn_r",
        **kwargs,
    ):
        """Example pipeline for comparing test data with itself (expanding test set)

        :param str hists_key: key to test histograms in datastore. default is 'test_hists'
        :param str time_axis: name of datetime feature. default is 'date' (column should be timestamp, date(time) or numeric batch id)
        :param int window: window size for trend detection. default is 10
        :param int shift: shift in expanding window. default is 1
        :param dict monitoring_rules: traffic light rules
        :param dict pull_rules: pull rules to determine dynamic boundaries
        :param list features: features of histograms to pick up from input data (optional)
        :param bool skip_empty_plots: if false, show empty plots in report with only nans or zeroes (optional)
        :param int last_n: plot statistic data for last 'n' periods (optional)
        :param int top_n: plot heatmap for top 'n' categories. default is 20 (optional)
        :param int plot_hist_n: plot histograms for last 'n' periods. default is 1 (optional)
        :param str report_filepath: the file path where to output the report (optional)
        :param list show_stats: list of statistic name patterns to show in the report. If None, show all (optional)
        :param list disable_heatmap: list of heatmap types to disable in the report. 'normal' to disable normal heatmap, 'row' to row normalized, 'column' to disable column normalized. If None, show all (optional)
        :param kwargs: residual keyword arguments
        :return: assembled expanding reference pipeline
        """
        modules = [
            ExpandingReferenceMetricsPipeline(
                hists_key,
                time_axis,
                window,
                shift,
                monitoring_rules,
                pull_rules,
                features,
                **kwargs,
            ),
            ReportPipe(
                sections_key="report_sections",
                store_key="html_report",
                skip_empty_plots=skip_empty_plots,
                last_n=last_n,
                top_n=top_n,
                plot_hist_n=plot_hist_n,
                report_filepath=report_filepath,
                show_stats=show_stats,
                disable_heatmap=disable_heatmap,
                cmap=cmap,
            ),
        ]

        super().__init__(modules)


class ReportPipe(Pipeline):
    """Pipeline of modules for generating sections and a final report."""

    def __init__(
        self,
        sections_key="report_sections",
        store_key="html_report",
        profiles_section="Profiles",
        comparisons_section="Comparisons",
        traffic_lights_section="Traffic Lights",
        alerts_section="Alerts",
        histograms_section="Histograms",
        report_filepath=None,
        show_stats=None,
        skip_empty_plots=True,
        last_n=0,
        top_n=20,
        skip_first_n=0,
        skip_last_n=0,
        plot_hist_n=6,
        disable_heatmap=None,
        cmap=None,
    ):
        """Initialize an instance of Report.

        :param str sections_key: key to store sections data in the datastore
        :param str store_key: key to store the HTML report data in the datastore
        :param str profiles_section: name for the profile data section. default is 'Profiles'
        :param str comparisons_section: name for the comparison data section. default is 'Comparisons'
        :param str traffic_lights_section: name for the traffic light section. default is 'Traffic Lights'
        :param str alerts_section: name for the alerts section. default is 'Alerts'
        :param str histograms_section: name for the histograms section. default is 'Histograms'
        :param str report_filepath: the file path where to output the report (optional)
        :param bool skip_empty_plots: if false, also show empty plots in report with only nans or zeroes (optional)
        :param int last_n: plot statistic data for last 'n' periods (optional)
        :param int top_n: plot heatmap for top 'n' categories. default is 20 (optional)
        :param int skip_first_n: when plotting data skip first 'n' periods. last_n takes precedence (optional)
        :param int skip_last_n: when plotting data skip last 'n' periods. last_n takes precedence (optional)
        :param int plot_hist_n: plot histograms for last 'n' periods. default is 1 (optional)
        :param list show_stats: list of statistic name patterns to show in the report. If None, show all (optional)
        :param list disable_heatmap: list of heatmap types to disable in the report. 'normal' to disable normal heatmap, 'row' to row normalized, 'column' to disable column normalized. If None, show all (optional)
        """
        self.store_key = store_key

        # dictionary of section descriptions
        descs = config["section_descriptions"]

        # default keyword arguments for each section
        sg_kws = {
            "store_key": sections_key,
            "skip_empty_plots": skip_empty_plots,
            "last_n": last_n,
            "skip_first_n": skip_first_n,
            "skip_last_n": skip_last_n,
            "show_stats": show_stats,
        }

        modules = [
            OverviewSectionGenerator(
                read_key="traffic_lights",
                description=descs.get("overview", ""),
                section_name="Overview",
                **sg_kws,
            ),
            # generate section with histogram
            HistogramSection(
                read_key="split_hists",
                store_key=sections_key,
                section_name=histograms_section,
                hist_name_starts_with="histogram",
                last_n=plot_hist_n,
                top_n=top_n,
                description=descs.get("histograms", ""),
                disable_heatmap=disable_heatmap,
                cmap=cmap,
            ),
            # section showing all traffic light alerts of monitored statistics
            TrafficLightSectionGenerator(
                read_key="traffic_lights",
                description=descs.get("traffic_lights", ""),
                section_name=traffic_lights_section,
                **sg_kws,
            ),
            # section with a summary of traffic light alerts
            AlertSectionGenerator(
                read_key="alerts",
                description=descs.get("alerts", ""),
                section_name=alerts_section,
                **sg_kws,
            ),
            # section of histogram and pull comparison statistics
            SectionGenerator(
                dynamic_bounds="dynamic_bounds_comparisons",
                static_bounds="static_bounds_comparisons",
                section_name=comparisons_section,
                ignore_stat_endswith=["_mean", "_std", "_pull"],
                read_key="comparisons",
                description=descs.get("comparisons", ""),
                **sg_kws,
            ),
            # section of profiled statistics with dynamic or static traffic light bounds
            SectionGenerator(
                dynamic_bounds="dynamic_bounds",
                section_name=profiles_section,
                static_bounds="static_bounds",
                ignore_stat_endswith=["_mean", "_std", "_pull"],
                read_key="profiles",
                description=descs.get("profiles", ""),
                **sg_kws,
            ),
            # generate report
            ReportGenerator(read_key=sections_key, store_key=store_key),
        ]
        if isinstance(report_filepath, (str, Path)) and len(report_filepath) > 0:
            modules.append(FileWriter(store_key, file_path=report_filepath))

        super().__init__(modules=modules)

    def transform(self, datastore):
        self.logger.info(f'Generating report "{self.store_key}".')
        return super().transform(datastore)
